"""
Constants used across package
"""

from enum import Enum


class Max(Enum):
    """Max number of elements on the panel"""

    AREAS = 8
    COUNTERS = 64
    KEYPADS = 16
    OUTPUTS = 208
    SETTINGS = 20
    TASKS = 32
    THERMOSTATS = 16
    USERS = 203
    LIGHTS = 256
    ZONES = 208
    ZONE_TEMPS = 16


class ZoneType(Enum):
    """Types of Elk zones"""

    DISABLED = 0
    BURGLAR_ENTRY_EXIT_1 = 1
    BURGLAR_ENTRY_EXIT_2 = 2
    BURGLAR_PERIMETER_INSTANT = 3
    BURGLAR_INTERIOR = 4
    BURGLAR_INTERIOR_FOLLOWER = 5
    BURGLAR_INTERIOR_NIGHT = 6
    BURGLAR_INTERIOR_NIGHT_DELAY = 7
    BURGLAR24_HOUR = 8
    BURGLAR_BOX_TAMPER = 9
    FIRE_ALARM = 10
    FIRE_VERIFIED = 11
    FIRE_SUPERVISORY = 12
    AUX_ALARM_1 = 13
    AUX_ALARM_2 = 14
    KEYFOB = 15
    NON_ALARM = 16
    CARBON_MONOXIDE = 17
    EMERGENCY_ALARM = 18
    FREEZE_ALARM = 19
    GAS_ALARM = 20
    HEAT_ALARM = 21
    MEDICAL_ALARM = 22
    POLICE_ALARM = 23
    POLICE_NO_INDICATION = 24
    WATER_ALARM = 25
    KEY_MOMENTARY_ARM_DISARM = 26
    KEY_MOMENTARY_ARM_AWAY = 27
    KEY_MOMENTARY_ARM_STAY = 28
    KEY_MOMENTARY_DISARM = 29
    KEY_ON_OFF = 30
    MUTE_AUDIBLES = 31
    POWER_SUPERVISORY = 32
    TEMPERATURE = 33
    ANALOG_ZONE = 34
    PHONE_KEY = 35
    INTERCOM_KEY = 36


class ZonePhysicalStatus(Enum):
    """Zone physical status."""

    UNCONFIGURED = 0
    OPEN = 1
    EOL = 2
    SHORT = 3


class ZoneLogicalStatus(Enum):
    """Zone logical status."""

    NORMAL = 0
    TROUBLED = 1
    VIOLATED = 2
    BYPASSED = 3


class ArmedStatus(Enum):
    """Area arming status: armed status"""

    DISARMED = "0"
    ARMED_AWAY = "1"
    ARMED_STAY = "2"
    ARMED_STAY_INSTANT = "3"
    ARMED_TO_NIGHT = "4"
    ARMED_TO_NIGHT_INSTANT = "5"
    ARMED_TO_VACATION = "6"


class ArmUpState(Enum):
    """Area arming status: Ability to arm"""

    NOT_READY_TO_ARM = "0"
    READY_TO_ARM = "1"
    CAN_BE_FORCE_ARMED = "2"
    ARMED_AND_EXIT_TIMER_RUNNING = "3"
    FULLY_ARMED = "4"
    FORCE_ARMED = "5"
    ARMED_WITH_BYPASS = "6"


class AlarmState(Enum):
    """Area arming status: Current alarm state"""

    NO_ALARM_ACTIVE = "0"
    ENTRANCE_DELAY_ACTIVE = "1"
    ALARM_ABORT_DELAY_ACTIVE = "2"
    FIRE_ALARM = "3"
    MEDICAL_ALARM = "4"
    POLICE_ALARM = "5"
    BURGLAR_ALARM = "6"
    AUX_1_ALARM = "7"
    AUX_2_ALARM = "8"
    AUX_3_ALARM = "9"
    AUX_4_ALARM = ":"
    CARBON_MONOXIDE_ALARM = ";"
    EMERGENCY_ALARM = "<"
    FREEZE_ALARM = "="
    GAS_ALARM = ">"
    HEAT_ALARM = "?"
    WATER_ALARM = "@"
    FIRE_SUPERVISORY = "A"
    VERIFY_FIRE = "B"


class ArmLevel(Enum):
    """Levels for Arm/Disarm al_encode messages"""

    DISARM = "0"
    ARMED_AWAY = "1"
    ARMED_STAY = "2"
    ARMED_STAY_INSTANT = "3"
    ARMED_NIGHT = "4"
    ARMED_NIGHT_INSTANT = "5"
    ARMED_VACATION = "6"
    ARM_TO_NEXT_AWAY_MODE = "7"
    ARM_TO_NEXT_STAY_MODE = "8"
    FORCE_ARM_TO_AWAY_MODE = "9"
    FORCE_ARM_TO_STAY_MODE = ":"


class ZoneAlarmState(Enum):
    """Alarm state for a zone"""

    NO_ALARM = "0"
    BURGLAR_ENTRY_EXIT_1 = "1"
    BURGLAR_ENTRY_EXIT_2 = "2"
    BURGLAR_PERIMETER_INSTANT = "3"
    BURGLAR_INTERIOR = "4"
    BURGLAR_INTERIOR_FOLLOWER = "5"
    BURGLAR_INTERIOR_NIGHT = "6"
    BURGLAR_INTERIOR_NIGHT_DELAY = "7"
    BURGLAR_24_HOUR = "8"
    BURGLAR_BOX_TAMPER = "9"
    FIRE_ALARM = ":"
    FIRE_VERIFIED = ";"
    FIRE_SUPERVISORY = "<"
    AUX_ALARM_1 = "="
    AUX_ALARM_2 = ">"
    KEYFOB = "?"  # not used
    NON_ALARM = "@"  # not used
    CARBON_MONOXIDE = "A"
    EMERGENCY_ALARM = "B"
    FREEZE_ALARM = "C"
    GAS_ALARM = "D"
    HEAT_ALARM = "E"
    MEDICAL_ALARM = "F"
    POLICE_ALARM = "G"
    POLICE_NO_INDICATION = "H"
    WATER_ALARM = "I"


class KeypadKeys(Enum):
    """Keys on the keypad."""

    USER_CODE_ENTERED = 0
    STAR = 11
    POUND = 12
    F1 = 13
    F2 = 14
    F3 = 15
    F4 = 16
    STAY = 17
    EXIT = 18
    CHIME = 19
    BYPASS = 20
    ELK = 21
    DOWN = 22
    UP = 23
    RIGHT = 24
    LEFT = 25
    F6 = 26
    F5 = 27
    DATA_KEY_MODE = 28


class TextDescriptions(Enum):
    """Types of description strings that can be retrieved from the panel"""

    ZONE = (0, Max.ZONES.value)
    AREA = (1, Max.AREAS.value)
    USER = (2, Max.USERS.value)
    KEYPAD = (3, Max.KEYPADS.value)
    OUTPUT = (4, 64)
    TASK = (5, Max.TASKS.value)
    TELEPHONE = 6
    LIGHT = (7, Max.LIGHTS.value)
    ALARM_DURATION = 8
    SETTING = (9, Max.SETTINGS.value)
    COUNTER = (10, Max.COUNTERS.value)
    THERMOSTAT = (11, Max.THERMOSTATS.value)
    FUNCTION_KEY_1 = 12
    FUNCTION_KEY_2 = 13
    FUNCTION_KEY_3 = 14
    FUNCTION_KEY_4 = 15
    FUNCTION_KEY_5 = 16
    FUNCTION_KEY_6 = 17
    AUDIO_ZONE = 18
    AUDIO_SOURCE = 19


# Map to convert message code to descriptive string
MESSAGE_MAP = {
    "AM": "Alarm memory update",
    "AP": "Send ASCII String",
    "AR": "Alarm Reporting to Ethernet",
    "AS": "Arming status report data",
    "AT": "Ethernet Test to IP",
    "AZ": "Alarm by zone reply",
    "CA": "Reply Touchscreen audio command",
    "CC": "Control output change update",
    "CD": "Outgoing Audio Equip Command",
    "CR": "Custom value report data",
    "CS": "Control output status report data",
    "CU": "Change user code reply",
    "CV": "Counter Value Data",
    "DK": "Display KP LCD Data, not used",
    "DS": "Lighting Poll Response",
    "EE": "Entry/Exit Time Data",
    "EM": "Email Trigger to M1XEP",
    "IC": "Send invalid user code digits",
    "IE": "Installer program exited",
    "IP": "M1XSP Insteon Program",
    "IR": "M1XSP Insteon Read",
    "KA": "Keypad areas report data",
    "KC": "Keypad key change update",
    "KF": "Function key pressed data",
    "LD": "Log data with index",
    "LW": "Reply temperature data",
    "NS": "Reply Source Name",
    "NZ": "Reply Zone Name",
    "PC": "PLC change update",
    "PS": "PLC status report data",
    "RE": "Reset Ethernet Module",
    "RP": "ELKRP connected",
    "RR": "Real Time Clock Data",
    "SD": "Text string description report data",
    "SS": "System Trouble Status data",
    "ST": "Temperature report data",
    "T2": "Reply Omnistat 2 data",
    "TC": "Task change update",
    "TR": "Thermostat data report",
    "UA": "User code areas report data",
    "VN": "Reply Version Number of M1",
    "XB": "reserved by ELKRP",
    "XK": "Request Ethernet test",
    "ZB": "Zone bypass report data",
    "ZC": "Zone change update",
    "ZD": "Zone definition report data",
    "ZP": "Zone partition report data",
    "ZS": "Zone status report data",
    "ZV": "Zone analog voltage data",
    "a0": "Disarm",
    "a1": "Arm to away",
    "a2": "Arm to stay",
    "a3": "Arm to stay instant",
    "a4": "Arm to night",
    "a5": "Arm to night instant",
    "a6": "Arm to vacation",
    "a7": "Arm, step to next Away Mode",
    "a8": "Arm, step to next Stay Mode",
    "a9": "Force Arm to Away Mode",
    "a:": "Force Arm to Stay Mode",
    "ar": "Alarm Reporting Acknowledge",
    "as": "Request arming status",
    "at": "Ethernet Test Acknowledge",
    "az": "Alarm by zone request",
    "ca": "Request Touchscreen audio command",
    "cd": "Incoming Audio Equip Command",
    "cf": "Control output OFF",
    "cn": "Control output ON",
    "cp": "Request ALL custom values",
    "cr": "Request custom value",
    "cs": "Control output status request",
    "ct": "Control output TOGGLE",
    "cu": "Change user code request",
    "cv": "Request Counter value",
    "cw": "Write custom value data",
    "cx": "Write counter value",
    "dm": "Display message",
    "ds": "Lighting Poll Request",
    "ip": "M1XSP Insteon Program",
    "ir": "M1XSP Insteon Read",
    "ka": "Request keypad areas",
    "kc": "Request F Key illumination status",
    "kf": "Request simulated function key press",
    "ld": "Request log data, with index",
    "le": "Write Log Data Entry",
    "lw": "Request temperature data",
    "pc": "Control any PLC device",
    "pf": "Turn OFF PLC device",
    "pn": "Turn ON PLC device",
    "ps": "Request PLC status",
    "pt": "Toggle PLC device",
    "rr": "Request Real Time Clock Read",
    "rs": "Used by Touchscreen",
    "rw": "Real Time Clock Write",
    "sd": "Request text string descriptions",
    "sp": "Speak phrase",
    "ss": "Request System Trouble Status",
    "st": "Request temperature",
    "sw": "Speak word",
    "t2": "Request Omnistat 2 data",
    "tn": "Task activation",
    "tr": "Request thermostat data",
    "ts": "Set thermostat data",
    "ua": "Request user code areas",
    "vn": "request Version Number of M1",
    "xk": "Reply from Ethernet test",
    "zb": "Zone bypass request",
    "zd": "Request zone definition data",
    "zp": "Zone partition request",
    "zs": "Zone status request",
    "zv": "Request Zone analog voltage",
}


class ThermostatSetting(Enum):
    """Thermostat consts when setting"""

    MODE = 0
    HOLD = 1
    FAN = 2
    GET_TEMPERATURE = 3
    COOL_SETPOINT = 4
    HEAT_SETPOINT = 5


class ThermostatMode(Enum):
    """Thermostat modes"""

    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    EMERGENCY_HEAT = 4


class ThermostatFan(Enum):
    """Thermostat fan"""

    AUTO = 0
    ON = 1


class ThermostatHold(Enum):
    """Thermostat hold"""

    OFF = 0
    ON = 1


class SettingFormat(Enum):
    """Types of values for settings."""

    NUMBER = 0
    TIMER = 1
    TIME_OF_DAY = 2


class ElkRPStatus(Enum):
    """Elk remote programming status."""

    DISCONNECTED = 0
    CONNECTED = 1
    INITIALIZING = 2



"""Definition of an ElkM1 Custom Value"""

from __future__ import annotations

from typing import Any, cast

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import cp_encode, cw_encode


class Setting(Element):
    """Class representing an Custom Value"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.value_format = 0
        self.value = None

    def set(self, value: int | tuple[int, int]) -> None:
        """(Helper) Set custom value."""
        self._connection.send(cw_encode(self._index, value, self.value_format))


class Settings(Elements):
    """Handling for multiple custom values"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Setting, Max.SETTINGS.value)
        connection.msg_decode.add_handler("CR", self._cr_handler)

    def sync(self) -> None:
        """Retrieve custom values from ElkM1"""
        self._connection.send(cp_encode())
        self.get_descriptions(TextDescriptions.SETTING.value)

    def _cr_handler(self, values: list[dict[str, Any]]) -> None:
        settings: list[Setting] = cast(list[Setting], self.elements)
        for value in values:
            setting = settings[value["index"]]
            setting.value_format = value["value_format"]
            setting.value = value["value"]



"""Discovery of Elk panels."""

from __future__ import annotations

import asyncio
import logging
import socket
import time
from collections.abc import Callable
from dataclasses import dataclass
from struct import unpack
from typing import Optional, cast

_LOGGER = logging.getLogger(__name__)


@dataclass
class ElkSystem:
    """An ELKM1 system."""

    mac_address: str
    ip_address: str
    port: int


def create_udp_socket(discovery_port: int) -> socket.socket:
    """Create a udp socket used for communicating with the device."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.bind(("", 0))
    sock.setblocking(False)
    return sock


class ELKDiscovery(asyncio.DatagramProtocol):
    """Discovery main class."""

    def __init__(
        self,
        destination: tuple[str, int],
        on_response: Callable[[bytes, tuple[str, int]], None],
    ) -> None:
        self.transport = None
        self.destination = destination
        self.on_response = on_response

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """Trigger on_response."""
        self.on_response(data, addr)

    def error_received(self, exc: Optional[Exception]) -> None:
        """Handle error."""
        _LOGGER.error("ELKDiscovery error: %s", exc)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Do nothing on connection lost."""


def _decode_data(raw_response: bytes) -> ElkSystem:
    """Decode an ELK discovery response packet."""
    remain = raw_response[5:]
    data = remain[:14]
    (
        mac1,
        mac2,
        mac3,
        mac4,
        mac5,
        mac6,
        ipv4_1,
        ipv4_2,
        ipv4_3,
        ipv4_4,
        port,
        _,
    ) = unpack("!6B4BHH", data)
    mac_address = ":".join(
        [
            format(val, "02x").upper().zfill(2)
            for val in (mac1, mac2, mac3, mac4, mac5, mac6)
        ]
    )
    ip_address = f"{ipv4_1}.{ipv4_2}.{ipv4_3}.{ipv4_4}"

    return ElkSystem(mac_address, ip_address, port)


class AIOELKDiscovery:
    """A 30303 discovery scanner."""

    DISCOVERY_PORT = 2362
    BROADCAST_FREQUENCY = 3
    DISCOVER_MESSAGE = b"XEPID"
    BROADCAST_ADDRESS = "<broadcast>"

    def __init__(self) -> None:
        self.found_devices: list[ElkSystem] = []

    def _destination_from_address(self, address: Optional[str]) -> tuple[str, int]:
        if address is None:
            address = self.BROADCAST_ADDRESS
        return (address, self.DISCOVERY_PORT)

    def _process_response(
        self,
        data: Optional[bytes],
        from_address: tuple[str, int],
        address: Optional[str],
        response_list: dict[tuple[str, int], ElkSystem],
    ) -> bool:
        """Process a response.

        Returns True if processing should stop
        """
        if (
            data is None
            or data == self.DISCOVER_MESSAGE
            or not data.startswith(b"M1XEP")
        ):
            return False
        try:
            response_list[from_address] = _decode_data(data)
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.warning("Failed to decode response from %s: %s", from_address, ex)
            return False
        return from_address[0] == address

    async def _async_run_scan(
        self,
        transport: asyncio.DatagramTransport,
        destination: tuple[str, int],
        timeout: int,
        found_all_future: "asyncio.Future[bool]",
    ) -> None:
        """Send the scans."""
        _LOGGER.debug("discover: %s => %s", destination, self.DISCOVER_MESSAGE)
        transport.sendto(self.DISCOVER_MESSAGE, destination)
        quit_time = time.monotonic() + timeout
        remain_time = float(timeout)
        while True:
            time_out = min(remain_time, timeout / self.BROADCAST_FREQUENCY)
            if time_out <= 0:
                return
            try:
                await asyncio.wait_for(
                    asyncio.shield(found_all_future), timeout=time_out
                )
            except asyncio.TimeoutError:
                if time.monotonic() >= quit_time:
                    return
                # No response, send broadcast again in cast it got lost
                _LOGGER.debug("discover: %s => %s", destination, self.DISCOVER_MESSAGE)
                transport.sendto(self.DISCOVER_MESSAGE, destination)
            else:
                return  # found_all
            remain_time = quit_time - time.monotonic()

    async def async_scan(
        self, timeout: int = 10, address: Optional[str] = None
    ) -> list[ElkSystem]:
        """Discover ELK devices."""
        sock = create_udp_socket(self.DISCOVERY_PORT)
        destination = self._destination_from_address(address)
        found_all_future: asyncio.Future[bool] = asyncio.Future()
        response_list: dict[tuple[str, int], ElkSystem] = {}

        def _on_response(data: bytes, addr: tuple[str, int]) -> None:
            _LOGGER.debug("discover: %s <= %s", addr, data)
            if self._process_response(data, addr, address, response_list):
                found_all_future.set_result(True)

        transport, _ = await asyncio.get_running_loop().create_datagram_endpoint(
            lambda: ELKDiscovery(
                destination=destination,
                on_response=_on_response,
            ),
            sock=sock,
        )
        try:
            await self._async_run_scan(
                cast(asyncio.DatagramTransport, transport),
                destination,
                timeout,
                found_all_future,
            )
        finally:
            transport.close()

        self.found_devices = list(response_list.values())
        return self.found_devices



"""Definition of an ElkM1 User"""

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements


class User(Element):
    """Class representing an User"""


class Users(Elements):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, User, Max.USERS.value)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self.get_descriptions(TextDescriptions.USER.value)

    def username(self, user_number: int) -> str:
        """Return name of user."""
        if 0 <= user_number < self.max_elements:
            return self.elements[user_number].name
        if user_number == 201:
            return "*Program*"
        if user_number == 202:
            return "*Elk RP*"
        if user_number == 203:
            return "*Quick arm*"
        return ""



"""Definition of an ElkM1 Zone"""

from .connection import Connection
from .const import (
    Max,
    TextDescriptions,
    ZoneLogicalStatus,
    ZonePhysicalStatus,
    ZoneType,
)
from .elements import Element, Elements
from .message import (
    az_encode,
    zb_encode,
    zd_encode,
    zp_encode,
    zs_encode,
    zt_encode,
    zv_encode,
)


class Zone(Element):
    """Class representing a Zone"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.definition = 0
        self.area = -1
        self.logical_status = 0
        self.physical_status = 0
        self.voltage = 0
        self.temperature = -60
        self.triggered_alarm = False

    def __str__(self) -> str:
        return (
            f"{self._index} '{self.name}' type:{ZoneType(self.definition).name} status:"
            f"{ZoneLogicalStatus(self.logical_status).name}/"
            f"{ZonePhysicalStatus(self.physical_status).name} "
            f"area:{self.area} trig:{self.triggered_alarm} "
            f"volt:{self.voltage} temp:{self.temperature}"
        )

    def bypass(self, code: int) -> None:
        """(Helper) Bypass zone."""
        self._connection.send(zb_encode(self._index, 0, code))

    def trigger(self) -> None:
        """(Helper) Trigger zone."""
        self._connection.send(zt_encode(self._index))

    def get_voltage(self) -> None:
        """(Helper) Get zone voltage."""
        self._connection.send(zv_encode(self._index))


class Zones(Elements):
    """Handling for multiple zones"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Zone, Max.ZONES.value)
        connection.msg_decode.add_handler("AZ", self._az_handler)
        connection.msg_decode.add_handler("LW", self._lw_handler)
        connection.msg_decode.add_handler("ST", self._st_handler)
        connection.msg_decode.add_handler("ZB", self._zb_handler)
        connection.msg_decode.add_handler("ZC", self._zc_handler)
        connection.msg_decode.add_handler("ZD", self._zd_handler)
        connection.msg_decode.add_handler("ZP", self._zp_handler)
        connection.msg_decode.add_handler("ZS", self._zs_handler)
        connection.msg_decode.add_handler("ZV", self._zv_handler)

    def sync(self) -> None:
        """Retrieve zones from ElkM1"""
        self._connection.send(az_encode())
        self._connection.send(zd_encode())
        self._connection.send(zp_encode())
        self._connection.send(zs_encode())
        self.get_descriptions(TextDescriptions.ZONE.value)

    def _az_handler(self, alarm_status: str) -> None:
        for zone in self.elements:
            zone.setattr("triggered_alarm", alarm_status[zone.index] != "0", True)

    def _lw_handler(self, keypad_temps: list[int], zone_temps: list[int]) -> None:
        for i in range(16):
            zone = self.elements[i]
            if zone_temps[zone.index] > -60:
                zone.setattr("temperature", zone_temps[zone.index], True)

    def _st_handler(self, group: int, device: int, temperature: int) -> None:
        if group == 0:
            self.elements[device].setattr("temperature", temperature, True)

    def _zb_handler(self, zone_number: int, zone_bypassed: bool) -> None:
        # If specific zone number was specified, then a ZC (zone change)
        # message will be received to update the bypass state.
        # If zone was 000 or 999 then we don't know which area was bypassed or
        # cleared and there is no ZC. Retrieve the current zone statuses...
        if zone_number < 0 or zone_number >= Max.ZONES.value:
            self._connection.send(zs_encode())

    def _zc_handler(self, zone_number: int, zone_status: tuple[int, int]) -> None:
        self.elements[zone_number].setattr("logical_status", zone_status[0], False)
        self.elements[zone_number].setattr("physical_status", zone_status[1], True)

    def _zd_handler(self, zone_definitions: list[int]) -> None:
        for zone in self.elements:
            zone.setattr("definition", zone_definitions[zone.index], True)

    def _zp_handler(self, zone_partitions: list[int]) -> None:
        for zone in self.elements:
            zone.setattr("area", zone_partitions[zone.index], True)

    def _zs_handler(self, zone_statuses: list[tuple[int, int]]) -> None:
        for zone in self.elements:
            zone.setattr("logical_status", zone_statuses[zone.index][0], False)
            zone.setattr("physical_status", zone_statuses[zone.index][1], True)

    def _zv_handler(self, zone_number: int, zone_voltage: float) -> None:
        self.elements[zone_number].setattr("voltage", zone_voltage, True)



"""Definition of an ElkM1 Area"""
import datetime as dt
import time
from typing import Any, Optional

from .connection import Connection
from .const import ElkRPStatus
from .elements import Element
from .message import lw_encode, rw_encode, sp_encode, ss_encode, sw_encode, vn_encode


class Panel(Element):
    """Class representing the overall Elk panel"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(0, connection)
        self.real_time_clock = None
        self.elkm1_version = None
        self.xep_version = None
        self.remote_programming_status = 0
        self.system_trouble_status = ""
        self.temperature_units = None
        self.user_code_length = None
        self._configured = True
        self.setattr("name", "ElkM1", True)

        self._connection.msg_decode.add_handler("VN", self._vn_handler)
        self._connection.msg_decode.add_handler("XK", self._xk_handler)
        self._connection.msg_decode.add_handler(
            "RR", self._xk_handler
        )  # RR/XK use same handler
        self._connection.msg_decode.add_handler("RP", self._rp_handler)
        self._connection.msg_decode.add_handler("SS", self._ss_handler)
        self._connection.msg_decode.add_handler("UA", self._ua_handler)

    def sync(self) -> None:
        """Retrieve panel information from ElkM1"""
        self._connection.send(vn_encode())
        self._connection.send(lw_encode())
        self._connection.send(ss_encode())
        # Don't sync UA from here as it is used as a "sync complete" marker

    def speak_word(self, word: int) -> None:
        """(Helper) Speak word."""
        self._connection.send(sw_encode(word))

    def speak_phrase(self, phrase: int) -> None:
        """(Helper) Speak phrase."""
        self._connection.send(sp_encode(phrase))

    def set_time(self, datetime: Optional[dt.datetime] = None) -> None:
        """(Helper) Set the time given a datetime."""
        if datetime is None:
            struct_time = time.localtime()
            datetime = dt.datetime(*struct_time[:6])
        self._connection.send(rw_encode(datetime))

    def _vn_handler(self, elkm1_version: str, xep_version: str) -> None:
        self.setattr("elkm1_version", elkm1_version, False)
        self.setattr("xep_version", xep_version, True)

    def _xk_handler(self, real_time_clock: str) -> None:
        self.setattr("real_time_clock", real_time_clock, True)

    def _ss_handler(self, system_trouble_status: str) -> None:
        def _get_status(index: int, trouble: str, zone_encoded: bool = False) -> None:
            if system_trouble_status[index] != "0":
                if zone_encoded:
                    zone = ord(system_trouble_status[index]) - 0x30
                    statuses.append(f"{trouble} zone {zone}")
                else:
                    statuses.append(trouble)

        statuses: list[str] = []
        _get_status(0, "AC Fail")
        _get_status(1, "Box Tamper", True)
        _get_status(2, "Fail To Communicate")
        _get_status(3, "EEProm Memory Error")
        _get_status(4, "Low Battery Control")
        _get_status(5, "Transmitter Low Battery", True)
        _get_status(6, "Over Current")
        _get_status(7, "Telephone Fault")
        _get_status(9, "Output 2")
        _get_status(10, "Missing Keypad")
        _get_status(11, "Zone Expander")
        _get_status(12, "Output Expander")
        _get_status(14, "ELKRP Remote Access")
        _get_status(16, "Common Area Not Armed")
        _get_status(17, "Flash Memory Error")
        _get_status(18, "Security Alert", True)
        _get_status(19, "Serial Port Expander")
        _get_status(20, "Lost Transmitter", True)
        _get_status(21, "GE Smoke CleanMe")
        _get_status(22, "Ethernet")
        _get_status(31, "Display Message In Keypad Line 1")
        _get_status(32, "Display Message In Keypad Line 2")
        _get_status(33, "Fire", True)
        self.setattr("system_trouble_status", ", ".join(statuses), True)

    def _rp_handler(self, remote_programming_status: int) -> None:
        if remote_programming_status == ElkRPStatus.DISCONNECTED.value:
            self._connection.resume()
        else:
            self._connection.pause()
        self.setattr("remote_programming_status", remote_programming_status, True)

    def _ua_handler(self, **kwargs: dict[str, Any]) -> None:
        self.setattr("user_code_length", kwargs["user_code_length"], False)
        self.setattr("temperature_units", kwargs["temperature_units"], True)



"""
ElkM1 message encode/decode.

Message format:
    LLMM<data>00CC, where:
    LL - length in ASCII hex
    MM - message code
    <data> - message contents; length varies by message code
    00 - constant two characters reserved (almost all messages)
    CC - checksum

The panel numbers zones, keypads, etc. in base 1. The interface of
this module refers to them using base 0. Conversion to base 1 is done
on sending and conversion to base 0 is done on receiving. So, base 0/1
conversion is encapsulated in this module.
"""

from __future__ import annotations

import datetime as dt
import re
import time
from collections import namedtuple
from collections.abc import Callable
from typing import Any, cast

from .const import Max

MessageEncode = namedtuple("MessageEncode", ["message", "response_command"])
# MsgHandler = Callable[[dict[str, Any]], None] # Gives type error
MsgHandler = Callable[..., None]

# pylint: disable=no-self-use
class MessageDecode:
    """Message decode and dispatcher."""

    def __init__(self) -> None:
        """Initialize a new Message instance."""
        self._handlers: dict[str, list[MsgHandler]] = {}

    def add_handler(self, message_type: str, handler: MsgHandler) -> None:
        """Add callback for handlers."""
        if message_type not in self._handlers:
            self._handlers[message_type] = []

        if handler not in self._handlers[message_type]:
            self._handlers[message_type].append(handler)

    def remove_handler(self, message_type: str, handler: MsgHandler) -> None:
        """Remove callback for handlers."""
        if message_type not in self._handlers:
            return
        if handler in self._handlers[message_type]:
            self._handlers[message_type].remove(handler)

    def call_handlers(self, cmd: str, decoded_msg: dict[str, Any]) -> None:
        """Call the message handlers."""
        # Copy the handlers list as add/remove could be
        # called when invoking the handlers
        handlers = list(self._handlers.get(cmd, []))

        for handler in handlers:
            handler(**decoded_msg)

    def decode(self, msg: str) -> None:
        """Decode an Elk message by passing to appropriate decoder"""
        valid, error_msg = _is_valid_length_and_checksum(msg)
        if valid:
            cmd = msg[2:4]
            decoder = getattr(self, f"_{cmd.lower()}_decode", None)
            if not decoder:
                cmd = "unknown"
                decoder = self._unknown_decode
            try:
                self.call_handlers(cmd, decoder(msg))
            except (IndexError, ValueError) as exc:
                raise ValueError("Cannot decode message") from exc
            return

        if not msg or msg.startswith("Username: ") or msg.startswith("Password: "):
            return
        if "Login successful" in msg:
            self.call_handlers("login", {"succeeded": True})
        elif msg.startswith("Username/Password not found") or msg == "Disabled":
            self.call_handlers("login", {"succeeded": False})
        else:
            raise ValueError(error_msg)

    def _am_decode(self, msg: str) -> dict[str, str]:
        """AM: Alarm memory by area report."""
        return {"alarm_memory": msg[4 : 4 + Max.AREAS.value]}

    def _as_decode(self, msg: str) -> dict[str, str]:
        """AS: Arming status report."""
        return {
            "armed_statuses": msg[4:12],
            "arm_up_states": msg[12:20],
            "alarm_states": msg[20:28],
        }

    def _az_decode(self, msg: str) -> dict[str, str]:
        """AZ: Alarm by zone report."""
        return {"alarm_status": msg[4 : 4 + Max.ZONES.value]}

    def _cr_one_custom_value_decode(self, index: int, part: str) -> dict[str, Any]:
        value = int(part[0:5])
        value_format = int(part[5])
        if value_format == 2:
            ret: int | tuple[int, int] = ((value >> 8) & 0xFF, value & 0xFF)
        else:
            ret = value
        return {"index": index, "value": ret, "value_format": value_format}

    def _cr_decode(self, msg: str) -> dict[str, Any]:
        """CR: Custom values"""
        if int(msg[4:6]) > 0:
            index = int(msg[4:6]) - 1
            return {"values": [self._cr_one_custom_value_decode(index, msg[6:12])]}

        part = 6
        ret = []
        for i in range(Max.SETTINGS.value):
            ret.append(self._cr_one_custom_value_decode(i, msg[part : part + 6]))
            part += 6
        return {"values": ret}

    def _cc_decode(self, msg: str) -> dict[str, Any]:
        """CC: Output status for single output."""
        return {"output": int(msg[4:7]) - 1, "output_status": msg[7] == "1"}

    def _cs_decode(self, msg: str) -> dict[str, Any]:
        """CS: Output status for all outputs."""
        output_status = [x == "1" for x in msg[4 : 4 + Max.OUTPUTS.value]]
        return {"output_status": output_status}

    def _cv_decode(self, msg: str) -> dict[str, Any]:
        """CV: Counter value."""
        return {"counter": int(msg[4:6]) - 1, "value": int(msg[6:11])}

    def _ee_decode(self, msg: str) -> dict[str, Any]:
        """EE: Entry/exit timer report."""
        return {
            "area": int(msg[4:5]) - 1,
            "is_exit": msg[5:6] == "0",
            "timer1": int(msg[6:9]),
            "timer2": int(msg[9:12]),
            "armed_status": msg[12:13],
        }

    def _ic_decode(self, msg: str) -> dict[str, Any]:
        """IC: Send Valid Or Invalid User Code Format."""
        code = msg[4:16]
        if re.match(r"(0\d){6}", code):
            code = re.sub(r"0(\d)", r"\1", code)
        return {
            "code": code,
            "user": int(msg[16:19]) - 1,
            "keypad": int(msg[19:21]) - 1,
        }

    def _ie_decode(self, msg: str) -> dict[str, str]:
        """IE: Installer mode exited."""
        return {}

    def _ka_decode(self, msg: str) -> dict[str, Any]:
        """KA: Keypad areas for all keypads."""
        return {"keypad_areas": [ord(x) - 0x31 for x in msg[4 : 4 + Max.KEYPADS.value]]}

    def _kc_decode(self, msg: str) -> dict[str, Any]:
        """KC: Keypad key change."""
        return {"keypad": int(msg[4:6]) - 1, "key": int(msg[6:8])}

    def _ld_decode(self, msg: str) -> dict[str, Any]:
        """LD: System Log Data Update."""
        area = int(msg[11]) - 1
        hour = int(msg[12:14])
        minute = int(msg[14:16])
        month = int(msg[16:18])
        day = int(msg[18:20])
        year = int(msg[24:26]) + 2000
        log_local_datetime = dt.datetime(year, month, day, hour, minute)
        log_local_time = time.mktime(log_local_datetime.timetuple())
        log_gm_timestruct = time.gmtime(log_local_time)

        log: dict[str, Any] = {}
        log["event"] = int(msg[4:8])
        log["number"] = int(msg[8:11])
        log["index"] = int(msg[20:23])
        log["timestamp"] = dt.datetime(
            *log_gm_timestruct[:6], tzinfo=dt.timezone.utc
        ).isoformat()

        return {"area": area, "log": log}

    def _lw_decode(self, msg: str) -> dict[str, Any]:
        """LW: temperatures from all keypads and zones 1-16."""
        keypad_temps = []
        zone_temps = []
        for i in range(16):
            keypad_temps.append(int(msg[4 + 3 * i : 7 + 3 * i]) - 40)
            zone_temps.append(int(msg[52 + 3 * i : 55 + 3 * i]) - 60)
        return {"keypad_temps": keypad_temps, "zone_temps": zone_temps}

    def _pc_decode(self, msg: str) -> dict[str, Any]:
        """PC: PLC (lighting) change."""
        housecode = msg[4:7]
        return {
            "housecode": housecode,
            "index": housecode_to_index(housecode),
            "light_level": int(msg[7:9]),
        }

    def _ps_decode(self, msg: str) -> dict[str, Any]:
        """PS: PLC (lighting) status."""
        return {
            "bank": ord(msg[4]) - 0x30,
            "statuses": [ord(x) - 0x30 for x in msg[5:69]],
        }

    def _rp_decode(self, msg: str) -> dict[str, int]:
        """RP: Remote programming status."""
        return {"remote_programming_status": int(msg[4:6])}

    def _rr_decode(self, msg: str) -> dict[str, str]:
        """RR: Realtime clock."""
        return {"real_time_clock": msg[4:20]}

    def _sd_decode(self, msg: str) -> dict[str, Any]:
        """SD: Description text."""
        desc_ch1 = msg[9]
        show_on_keypad = ord(desc_ch1) >= 0x80
        if show_on_keypad:
            desc_ch1 = chr(ord(desc_ch1) & 0x7F)
        return {
            "desc_type": int(msg[4:6]),
            "unit": int(msg[6:9]) - 1,
            "desc": (desc_ch1 + msg[10:25]).rstrip(),
            "show_on_keypad": show_on_keypad,
        }

    def _ss_decode(self, msg: str) -> dict[str, str]:
        """SS: System status."""
        return {"system_trouble_status": msg[4:-2]}

    def _st_decode(self, msg: str) -> dict[str, int]:
        """ST: Temperature update."""
        group = int(msg[4:5])
        temperature = int(msg[7:10])
        if group == 0:
            temperature -= 60
        elif group == 1:
            temperature -= 40
        return {"group": group, "device": int(msg[5:7]) - 1, "temperature": temperature}

    def _tc_decode(self, msg: str) -> dict[str, int]:
        """TC: Task change."""
        return {"task": int(msg[4:7]) - 1}

    def _tr_decode(self, msg: str) -> dict[str, Any]:
        """TR: Thermostat data response."""
        return {
            "thermostat_index": int(msg[4:6]) - 1,
            "mode": int(msg[6]),
            "hold": msg[7] == "1",
            "fan": int(msg[8]),
            "current_temp": int(msg[9:11]),
            "heat_setpoint": int(msg[11:13]),
            "cool_setpoint": int(msg[13:15]),
            "humidity": int(msg[15:17]),
        }

    def _ua_decode(self, msg: str) -> dict[str, Any]:
        """UA: Valid User Code Areas."""
        return {
            "user_code": int(msg[4:10]),
            "valid_areas": int(msg[10:12], 16),
            "diagnostic": msg[12:20],
            "user_code_length": int(msg[20]),
            "user_code_type": int(msg[21]),
            "temperature_units": msg[22],
        }

    def _vn_decode(self, msg: str) -> dict[str, str]:
        """VN: Version information."""
        elkm1_version = f"{int(msg[4:6], 16)}.{int(msg[6:8], 16)}.{int(msg[8:10], 16)}"
        xep_version = (
            f"{int(msg[10:12], 16)}.{int(msg[12:14], 16)}.{int(msg[14:16], 16)}"
        )
        return {"elkm1_version": elkm1_version, "xep_version": xep_version}

    def _xk_decode(self, msg: str) -> dict[str, str]:
        """XK: Ethernet Test."""
        return {"real_time_clock": msg[4:20]}

    def _zb_decode(self, msg: str) -> dict[str, Any]:
        """ZB: Zone bypass report."""
        return {"zone_number": int(msg[4:7]) - 1, "zone_bypassed": msg[7] == "1"}

    def _zc_decode(self, msg: str) -> dict[str, Any]:
        """ZC: Zone Change."""
        status = _status_decode(int(msg[7:8], 16))
        return {"zone_number": int(msg[4:7]) - 1, "zone_status": status}

    def _zd_decode(self, msg: str) -> dict[str, list[int]]:
        """ZD: Zone definitions."""
        zone_definitions = [ord(x) - 0x30 for x in msg[4 : 4 + Max.ZONES.value]]
        return {"zone_definitions": zone_definitions}

    def _zp_decode(self, msg: str) -> dict[str, list[int]]:
        """ZP: Zone partitions."""
        zone_partitions = [ord(x) - 0x31 for x in msg[4 : 4 + Max.ZONES.value]]
        return {"zone_partitions": zone_partitions}

    def _zs_decode(self, msg: str) -> dict[str, list[tuple[int, int]]]:
        """ZS: Zone statuses."""
        status = [_status_decode(int(x, 16)) for x in msg[4 : 4 + Max.ZONES.value]]
        return {"zone_statuses": status}

    def _zv_decode(self, msg: str) -> dict[str, Any]:
        """ZV: Zone voltage."""
        return {"zone_number": int(msg[4:7]) - 1, "zone_voltage": int(msg[7:10]) / 10}

    def _unknown_decode(self, msg: str) -> dict[str, str]:
        """Generic handler called when no specific handler exists"""
        return {"msg_code": msg[2:4], "data": msg[4:-2]}


# pylint: enable=no-self-use


def housecode_to_index(housecode: str) -> int:
    """Convert a X10 housecode to a zero-based index"""
    match = re.search(r"^([A-P])(\d{1,2})$", housecode.upper())
    if match:
        house_index = int(match.group(2))
        if 1 <= house_index <= 16:
            return (ord(match.group(1)) - ord("A")) * 16 + house_index - 1
    raise ValueError(f"Invalid X10 housecode: {housecode}")


def index_to_housecode(index: int) -> str:
    """Convert a zero-based index to a X10 housecode."""
    if index < 0 or index > 255:
        raise ValueError
    quotient, remainder = divmod(index, 16)
    return f"{chr(ord('A') + quotient)}{remainder + 1:02}"


def get_elk_command(line: str) -> str:
    """Return the 2 character command in the message."""
    if len(line) < 4:
        return ""
    return line[2:4]


def _status_decode(status: int) -> tuple[int, int]:
    """Decode a 1 byte status into logical and physical statuses."""
    logical_status = (status & 0b00001100) >> 2
    physical_status = status & 0b00000011
    return (logical_status, physical_status)


def _is_valid_length_and_checksum(msg: str) -> tuple[bool, str]:
    """Check packet length valid and that checksum is good."""
    try:
        if int(msg[:2], 16) != (len(msg) - 2):
            return False, "Incorrect message length"

        checksum = int(msg[-2:], 16)
        for char in msg[:-2]:
            checksum += ord(char)
        if (checksum % 256) != 0:
            return False, "Bad checksum"
    except ValueError:
        return False, "Message invalid"

    return True, ""


def al_encode(arm_mode: str, area: int, user_code: int) -> MessageEncode:
    """al: Arm system. Note in 'al' the 'l' can vary"""
    return MessageEncode(f"0Da{arm_mode}{area + 1:1}{user_code:06}00", "AS")


def as_encode() -> MessageEncode:
    """as: Get area status."""
    return MessageEncode("06as00", "AS")


def az_encode() -> MessageEncode:
    """az: Get alarm by zone."""
    return MessageEncode("06az00", "AZ")


def cf_encode(output: int) -> MessageEncode:
    """cf: Turn off output."""
    return MessageEncode(f"09cf{output + 1:03}00", None)


def ct_encode(output: int) -> MessageEncode:
    """ct: Toggle output."""
    return MessageEncode(f"09ct{output + 1:03}00", None)


def cn_encode(output: int, seconds: int) -> MessageEncode:
    """cn: Turn on output."""
    return MessageEncode(f"0Ecn{output + 1:03}{seconds:05}00", None)


def cs_encode() -> MessageEncode:
    """cs: Get all output status."""
    return MessageEncode("06cs00", "CS")


def cp_encode() -> MessageEncode:
    """cp: Get ALL custom values."""
    return MessageEncode("06cp00", "CR")


def cr_encode(index: int) -> MessageEncode:
    """cr: Get a custom value."""
    return MessageEncode(f"08cr{index + 1:02}00", "CR")


def cw_encode(
    index: int, value: int | tuple[int, int], value_format: int
) -> MessageEncode:
    """cw: Write a custom value."""
    if value_format == 2:
        val = cast(tuple[int, int], value)
        enc = val[0] * 256 + val[1]
    else:
        enc = cast(int, value)
    return MessageEncode(f"0Dcw{index + 1:02}{enc:05}00", None)


def cv_encode(counter: int) -> MessageEncode:
    """cv: Get counter."""
    return MessageEncode(f"08cv{counter + 1:02}00", "CV")


def cx_encode(counter: int, value: int) -> MessageEncode:
    """cx: Change counter value."""
    return MessageEncode(f"0Dcx{counter + 1:02}{value:05}00", "CV")


def dm_encode(
    keypad_area: int, clear: int, beep: bool, timeout: int, line1: str, line2: str
) -> MessageEncode:
    """dm: Display message on keypad."""
    return MessageEncode(
        f"2Edm{keypad_area + 1:1}{clear:1}{beep:1}{timeout:05}{line1:^<16.16}{line2:^<16.16}00",
        None,
    )


def ka_encode() -> MessageEncode:
    """ka: Get keypad areas."""
    return MessageEncode("06ka00", "KA")


def lw_encode() -> MessageEncode:
    """lw: Get temperature data."""
    return MessageEncode("06lw00", "LW")


def pc_encode(
    index: int, function_code: int, extended_code: int, seconds: int
) -> MessageEncode:
    """pc: Control any PLC device."""
    return MessageEncode(
        f"11pc{index_to_housecode(index)}{function_code:02}{extended_code:02}{seconds:04}00",
        None,
    )


def pf_encode(index: int) -> MessageEncode:
    """pf: Turn off light."""
    return MessageEncode(f"09pf{index_to_housecode(index)}00", None)


def pn_encode(index: int) -> MessageEncode:
    """pn: Turn on light."""
    return MessageEncode(f"09pn{index_to_housecode(index)}00", None)


def ps_encode(bank: int) -> MessageEncode:
    """ps: Get lighting status."""
    return MessageEncode(f"07ps{bank:1}00", "PS")


def pt_encode(index: int) -> MessageEncode:
    """pt: Toggle light."""
    return MessageEncode(f"09pt{index_to_housecode(index)}00", None)


def sd_encode(desc_type: int, unit: int) -> MessageEncode:
    """sd: Get description."""
    return MessageEncode(f"0Bsd{desc_type:02}{unit + 1:03}00", "SD")


def sp_encode(phrase: int) -> MessageEncode:
    """sp: Speak phrase."""
    return MessageEncode(f"09sp{phrase:03}00", None)


def ss_encode() -> MessageEncode:
    """ss: Get system trouble status."""
    return MessageEncode("06ss00", "SS")


def sw_encode(word: int) -> MessageEncode:
    """sw: Speak word."""
    return MessageEncode(f"09sw{word:03}00", None)


def rw_encode(date_time: dt.datetime) -> MessageEncode:
    """rw: Write time given a datetime."""
    elk_weekday = (date_time.weekday() + 1) % 7 + 1
    time_str = date_time.strftime("%S%M%H.%d%m%y").replace(".", str(elk_weekday))
    return MessageEncode(f"13rw{time_str}00", None)


def tn_encode(task: int) -> MessageEncode:
    """tn: Activate task."""
    return MessageEncode(f"09tn{task + 1:03}00", None)


def tr_encode(thermostat: int) -> MessageEncode:
    """tr: Request thermostat data."""
    return MessageEncode(f"08tr{thermostat + 1:02}00", None)


def ts_encode(thermostat: int, value: int, element: int) -> MessageEncode:
    """ts: Set thermostat data."""
    return MessageEncode(f"0Bts{thermostat + 1:02}{value:02}{element:1}00", None)


def ua_encode(user_code: int) -> MessageEncode:
    """ua: Requst valid user code areas"""
    return MessageEncode(f"0Cua{user_code:06}00", "UA")


def vn_encode() -> MessageEncode:
    """zd: Get panel software version information."""
    return MessageEncode("06vn00", "VN")


def zb_encode(zone: int, area: int, user_code: int) -> MessageEncode:
    """zb: Zone bypass. Zone < 0 unbypass all; Zone > Max bypass all."""
    if zone < 0:
        zone = 0
    elif zone > Max.ZONES.value:
        zone = 999
    else:
        zone += 1
    return MessageEncode(
        f"10zb{zone:03}{area + 1:1}{user_code:06}00",
        "ZB",
    )


def zd_encode() -> MessageEncode:
    """zd: Get zone definitions"""
    return MessageEncode("06zd00", "ZD")


def zp_encode() -> MessageEncode:
    """zp: Get zone partitions"""
    return MessageEncode("06zp00", "ZP")


def zs_encode() -> MessageEncode:
    """zs: Get zone statuses"""
    return MessageEncode("06zs00", "ZS")


def zt_encode(zone: int) -> MessageEncode:
    """zt: Trigger zone."""
    return MessageEncode(f"09zt{zone + 1:03}00", None)


def zv_encode(zone: int) -> MessageEncode:
    """zv: Get zone voltage"""
    return MessageEncode(f"09zv{zone + 1:03}00", "ZV")



"""
  Base of all the elements found on the Elk panel... Zone, Keypad, etc.
"""

from __future__ import annotations

import re
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, Generator, Type

from .connection import Connection
from .const import TextDescriptions
from .message import sd_encode


class Element:
    """Element class"""

    def __init__(self, index: int, connection: Connection) -> None:
        self._index = index
        self._connection = connection
        self._callbacks: list[Callable[[Element, dict[str, Any]], None]] = []
        self.name: str = self.default_name()
        self._changeset: dict[str, Any] = {}
        self._configured: bool = False

    @property
    def index(self) -> int:
        """Get the index, immutable once class created"""
        return self._index

    @property
    def configured(self) -> bool:
        """If a callback has ever been triggered this will be true."""
        return self._configured

    def add_callback(self, callback: Callable[[Element, dict[str, Any]], None]) -> None:
        """Callbacks when attribute of element changes"""
        self._callbacks.append(callback)

    def remove_callback(
        self, callback: Callable[[Element, dict[str, Any]], None]
    ) -> None:
        """Callbacks when attribute of element changes"""
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _call_callbacks(self) -> None:
        """Callbacks when attribute of element changes"""
        for callback in self._callbacks:
            callback(self, self._changeset)
        self._changeset = {}

    def setattr(
        self, attr: str, new_value: Any, close_the_changeset: bool = True
    ) -> None:
        """If attribute value has changed then set it and call the callbacks"""
        existing_value: Any = getattr(self, attr, None)
        if existing_value != new_value:
            setattr(self, attr, new_value)
            self._changeset[attr] = new_value

        if close_the_changeset and self._changeset:
            self._call_callbacks()

    def default_name(self, separator: str = "-") -> str:
        """Return a default name for based on class and index of element"""
        return f"{self.__class__.__name__}{separator}{self._index + 1:03}"

    def is_default_name(self) -> bool:
        """Check if the name assigned is the default_name"""
        return self.name == self.default_name()

    def __str__(self) -> str:
        varlist = {
            k: v
            for (k, v) in vars(self).items()
            if not k.startswith("_") and k != "name"
        }.items()
        varstr = " ".join(
            "%s:%s" % item  # pylint: disable=consider-using-f-string
            for item in varlist
        )
        return f"{self._index} '{self.name}' {varstr}"

    def as_dict(self) -> dict[str, Any]:
        """Package up the public attributes as a dict."""
        attrs = vars(self)
        return {key: attrs[key] for key in attrs if not key.startswith("_")}


class Elements:
    """Base for list of elements."""

    def __init__(
        self, connection: Connection, class_: Type[Element], max_elements: int
    ) -> None:
        self._connection = connection
        self.max_elements = max_elements
        self.elements = [class_(i, connection) for i in range(max_elements)]

        self._get_description_state: tuple[
            int, int, list[str | None], Callable[[list[str | None], int], None]
        ] | None = None
        connection.msg_decode.add_handler("SD", self._sd_handler)

    def __iter__(self) -> Generator[Element, None, None]:
        for element in self.elements:
            yield element

    def __getitem__(self, key: int) -> Element:
        return self.elements[key]

    def _sd_handler(
        self, desc_type: int, unit: int, desc: str, show_on_keypad: bool
    ) -> None:
        if not self._get_description_state:
            return
        (_desc_type, count, results, callback) = self._get_description_state
        if desc_type != _desc_type:
            return

        if unit < 0 or unit >= count:
            callback(results, desc_type)
            self._get_description_state = None
        else:
            results[unit] = desc
            self._connection.send(sd_encode(desc_type, unit + 1))

    def _got_desc(self, descriptions: list[str | None], desc_type: int) -> None:
        # Elk reports descriptions for all 199 users, irregardless of how many
        # are configured. Only set configured for those that are really there.
        if desc_type == TextDescriptions.USER.value[0]:
            user_re = re.compile(r"USER \d\d\d")
        else:
            user_re = None

        for element in self.elements:
            if element.index >= len(descriptions):
                break
            name = descriptions[element.index]
            if name is not None:
                if user_re and user_re.match(name):
                    continue
                element.setattr("name", name, True)
                element._configured = True  # pylint: disable=protected-access

    def get_descriptions(self, description_type: tuple[int, int]) -> None:
        """Gets the descriptions for specified type."""
        (desc_type, count) = description_type
        results: list[str | None] = [None] * count
        self._get_description_state = (desc_type, count, results, self._got_desc)
        self._connection.send(sd_encode(desc_type, 0))

    @abstractmethod
    def sync(self) -> None:
        """Synchronize elements"""



"""Definition of an ElkM1 Custom Value"""

from __future__ import annotations

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import cv_encode, cx_encode


class Counter(Element):
    """Class representing an Counter"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.value = None

    def get(self) -> None:
        """(Helper) Get counter"""
        self._connection.send(cv_encode(self._index))

    def set(self, value: int) -> None:
        """(Helper) Set counter to value"""
        self._connection.send(cx_encode(self._index, value))


class Counters(Elements):
    """Handling for multiple counters"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Counter, Max.COUNTERS.value)
        connection.msg_decode.add_handler("CV", self._cv_handler)

    def sync(self) -> None:
        """Retrieve values from ElkM1 on demand"""
        self.get_descriptions(TextDescriptions.COUNTER.value)

    def _got_desc(self, descriptions: list[str | None], desc_type: int) -> None:
        super()._got_desc(descriptions, desc_type)
        # Only poll counters that have a name defined
        for counter in self.elements:
            if not counter.is_default_name():
                self._connection.send(cv_encode(counter.index))

    def _cv_handler(self, counter: int, value: int) -> None:
        self.elements[counter].setattr("value", value, True)



"""Definition of an ElkM1 Thermostat"""

from __future__ import annotations

from .connection import Connection
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import tr_encode, ts_encode


class Thermostat(Element):
    """Class representing an Thermostat"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.mode = 0
        self.hold = False
        self.fan = 0
        self.current_temp = 0
        self.heat_setpoint = 0
        self.cool_setpoint = 0
        self.humidity = 0

    def set(self, element_to_set: int, value: int) -> None:
        """(Helper) Set thermostat"""
        self._connection.send(ts_encode(self.index, value, element_to_set))


class Thermostats(Elements):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Thermostat, Max.THERMOSTATS.value)
        connection.msg_decode.add_handler("ST", self._st_handler)
        connection.msg_decode.add_handler("TR", self._tr_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self.get_descriptions(TextDescriptions.THERMOSTAT.value)

    def _got_desc(self, descriptions: list[str | None], desc_type: int) -> None:
        super()._got_desc(descriptions, desc_type)
        # Only poll thermostats that have a name defined
        for thermostat in self.elements:
            if not thermostat.is_default_name():
                self._connection.send(tr_encode(thermostat.index))

    def _st_handler(self, group: int, device: int, temperature: int) -> None:
        if group == 2:
            self.elements[device].setattr("current_temp", temperature, True)

    def _tr_handler(
        self,
        thermostat_index: int,
        mode: int,
        hold: bool,
        fan: int,
        current_temp: int,
        heat_setpoint: int,
        cool_setpoint: int,
        humidity: int,
    ) -> None:
        thermostat = self.elements[thermostat_index]
        thermostat.setattr("mode", mode, False)
        thermostat.setattr("hold", hold, False)
        thermostat.setattr("fan", fan, False)
        thermostat.setattr("current_temp", current_temp, False)
        thermostat.setattr("heat_setpoint", heat_setpoint, False)
        thermostat.setattr("cool_setpoint", cool_setpoint, False)
        thermostat.setattr("humidity", humidity, True)



"""Async IO."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable
from functools import partial, reduce
from typing import Any, Optional, cast

import serial_asyncio

from .message import MessageDecode, MessageEncode, get_elk_command
from .util import parse_url, url_scheme_is_secure

LOG = logging.getLogger(__name__)


class Connection:
    """Method to manage a connection to the ElkM1."""

    def __init__(
        self, config: dict[str, Any], loop: asyncio.AbstractEventLoop | None = None
    ):
        """Setup a connection."""
        self._config = config
        self._loop = loop if loop else asyncio.get_event_loop()

        self._elk_protocol: _ElkProtocol | None = None
        self._msg_decode: MessageDecode = MessageDecode()
        self._connection_retry_time = 1
        self._reconnect_task: asyncio.TimerHandle | None = None

    async def connect(self) -> None:
        """Asyncio connection to Elk."""

        LOG.info("Connecting to ElkM1 at %s", self._config["url"])
        scheme, dest, param, ssl_context = parse_url(self._config["url"])
        heartbeat_time = 120 if scheme != "serial" else -1

        connection = partial(
            _ElkProtocol,
            self._loop,
            heartbeat_time,
            self._connected_callback,
            self._disconnected_callback,
            self._got_data_callback,
            self._timeout_callback,
        )
        try:
            if scheme == "serial":
                await serial_asyncio.create_serial_connection(
                    self._loop, connection, dest, baudrate=param
                )
            else:
                await asyncio.wait_for(
                    self._loop.create_connection(
                        connection, host=dest, port=param, ssl=ssl_context
                    ),
                    timeout=30,
                )
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            if self._connection_retry_time <= 0:
                return

            LOG.warning(
                "Could not connect to ElkM1 (%s). Retrying in %d seconds",
                err,
                self._connection_retry_time,
            )
            self._start_connection_retry_timer()

    def is_connected(self) -> bool:
        """Is the connection active?"""
        return self._elk_protocol is not None

    @property
    def msg_decode(self) -> MessageDecode:
        """Get the message decoder instance."""
        return self._msg_decode

    def send(self, msg: MessageEncode) -> None:
        """Send a message on the connection."""
        if self._elk_protocol:
            self._elk_protocol.write_data(msg.message, msg.response_command)

    def pause(self) -> None:
        """Pause the connection from sending/receiving."""
        if self._elk_protocol:
            self._elk_protocol.pause()

    def resume(self) -> None:
        """Restart the connection from sending/receiving."""
        if self._elk_protocol:
            self._elk_protocol.resume()

    def disconnect(self) -> None:
        """Disconnect the connection from sending/receiving."""
        self._connection_retry_time = -1  # Stop future timeout reconnects
        if self._elk_protocol is not None:
            self._elk_protocol.close()
            self._elk_protocol = None
        if self._reconnect_task is not None:
            self._reconnect_task.cancel()
            self._reconnect_task = None

    def _start_connection_retry_timer(self) -> None:
        timeout = self._connection_retry_time
        if timeout > 0:
            self._reconnect_task = self._loop.call_later(timeout, self._reconnect)
            self._connection_retry_time = min(60, timeout * 2)

    def _connected_callback(self, elk_protocol: _ElkProtocol) -> None:
        """Login and sync the ElkM1 panel to memory."""
        LOG.info("Connected to ElkM1")
        self._elk_protocol = elk_protocol
        self._connection_retry_time = 1
        self._msg_decode.call_handlers("connected", {})
        if elk_protocol is not None and url_scheme_is_secure(self._config["url"]):
            self._elk_protocol.write_data(self._config["userid"], raw=True)
            self._elk_protocol.write_data(self._config["password"], raw=True)
        else:
            self._msg_decode.call_handlers("login", {"succeeded": True})

    def _reconnect(self) -> None:
        asyncio.ensure_future(self.connect())

    def _disconnected_callback(self) -> None:
        LOG.warning("ElkM1 at %s disconnected", self._config["url"])
        self._elk_protocol = None
        self._start_connection_retry_timer()
        self._msg_decode.call_handlers("disconnected", {})

    def _got_data_callback(self, data: str) -> None:
        LOG.debug("got_data '%s'", data)
        try:
            self._msg_decode.decode(data)
        except (ValueError, AttributeError) as exc:
            LOG.error("Invalid message '%s'", data, exc_info=exc)

    def _timeout_callback(self, msg_code: str) -> None:
        self._msg_decode.call_handlers("timeout", {"msg_code": msg_code})


class _ElkProtocol(asyncio.Protocol):
    """asyncio Protocol with line parsing and queuing writes"""

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        heartbeat_time: float,
        connected: Callable[[_ElkProtocol], None],
        disconnected: Callable[[], None],
        got_data: Callable[[str], None],
        timeout: Callable[[str | None], None],
    ) -> None:
        self._loop = loop
        self._heartbeat_time = heartbeat_time
        self._connected_callback = connected
        self._disconnected_callback = disconnected
        self._got_data_callback = got_data
        self._timeout_callback = timeout

        self._transport: Optional[asyncio.Transport] = None
        self._waiting_for_response: Optional[str] = None
        self._write_timeout_task: Optional[asyncio.TimerHandle] = None
        self._heartbeat_timeout_task: Optional[asyncio.TimerHandle] = None
        self._queued_writes: list[tuple[str, Optional[str], float]] = []
        self._buffer = ""
        self._paused = False

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        LOG.debug("connected callback")
        self._transport = cast(asyncio.Transport, transport)
        self._connected_callback(self)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        LOG.debug("disconnected callback")
        self._transport = None
        self._cleanup()
        if self._disconnected_callback:
            self._disconnected_callback()

    def _cleanup(self) -> None:
        self._cancel_write_timer()
        self._cancel_heartbeat_timer()
        self._waiting_for_response = None
        self._queued_writes = []
        self._buffer = ""

    def close(self) -> None:
        """Stop the connection from sending/receiving/reconnecting."""
        if self._transport:
            self._transport.close()
            self._transport = None
        self._cleanup()

    def pause(self) -> None:
        """Pause the connection from sending/receiving."""
        self._cleanup()
        self._paused = True

    def resume(self) -> None:
        """Restart the connection from sending/receiving."""
        self._paused = False

    def _response_required_timeout(self) -> None:
        self._timeout_callback(self._waiting_for_response)
        self._write_timeout_task = None
        self._waiting_for_response = None
        self._process_write_queue()

    def _cancel_write_timer(self) -> None:
        if self._write_timeout_task:
            self._write_timeout_task.cancel()
            self._write_timeout_task = None

    def _cancel_heartbeat_timer(self) -> None:
        if self._heartbeat_timeout_task is not None:
            self._heartbeat_timeout_task.cancel()

    def _heartbeat_timeout(self) -> None:
        LOG.warning("ElkM1 connection heartbeat timed out, disconnecting")
        if self._transport:
            self._transport.close()

    def _restart_heartbeat_timer(self) -> None:
        self._cancel_heartbeat_timer()
        if self._heartbeat_time > 0:
            self._heartbeat_timeout_task = self._loop.call_later(
                self._heartbeat_time, self._heartbeat_timeout
            )

    def data_received(self, data: bytes) -> None:
        self._restart_heartbeat_timer()
        self._buffer += data.decode("ISO-8859-1")
        while "\r\n" in self._buffer:
            line, self._buffer = self._buffer.split("\r\n", 1)
            if get_elk_command(line) == self._waiting_for_response:
                self._waiting_for_response = None
                self._cancel_write_timer()
            self._got_data_callback(line)
        self._process_write_queue()

    def _process_write_queue(self) -> None:
        while self._queued_writes and not self._waiting_for_response:
            to_write = self._queued_writes.pop(0)
            self.write_data(to_write[0], to_write[1], timeout=to_write[2])

    def write_data(
        self,
        data: str,
        response_required: str | None = None,
        timeout: float = 5.0,
        raw: bool = False,
    ) -> None:
        """Write data on the asyncio Protocol"""
        if self._transport is None:
            return

        if self._paused:
            return

        if self._waiting_for_response:
            LOG.debug("queueing write %s", data)
            self._queued_writes.append((data, response_required, timeout))
            return

        if response_required:
            self._waiting_for_response = response_required
            if timeout > 0:
                self._write_timeout_task = self._loop.call_later(
                    timeout, self._response_required_timeout
                )

        if not raw:
            cksum = 256 - reduce(lambda x, y: x + y, map(ord, data)) % 256
            data = f"{data}{cksum:02X}"
            if int(data[0:2], 16) != len(data) - 2:
                LOG.debug("message length wrong: %s", data)

        LOG.debug("write_data '%s'", data)
        self._transport.write((data + "\r\n").encode())



"""Definition of an ElkM1 Keypad."""
import datetime as dt
from typing import Optional, cast

from .connection import Connection
from .const import KeypadKeys, Max, TextDescriptions
from .elements import Element, Elements
from .message import ka_encode


class Keypad(Element):
    """Class representing an Keypad"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.area = -1
        self.temperature = -40
        self.last_user_time = dt.datetime.now(dt.timezone.utc)
        self.last_user = -1
        self.code = ""
        self.last_keypress: Optional[tuple[str, int]] = None


class Keypads(Elements):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Keypad, Max.KEYPADS.value)
        connection.msg_decode.add_handler("IC", self._ic_handler)
        connection.msg_decode.add_handler("KA", self._ka_handler)
        connection.msg_decode.add_handler("KC", self._kc_handler)
        connection.msg_decode.add_handler("LW", self._lw_handler)
        connection.msg_decode.add_handler("ST", self._st_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self._connection.send(ka_encode())
        self.get_descriptions(TextDescriptions.KEYPAD.value)

    def _ic_handler(self, code: int, user: int, keypad: int) -> None:
        keypad_ = self.elements[keypad]

        # By setting a time this will force the IC change to always be reported
        keypad_.setattr("last_user_time", dt.datetime.now(dt.timezone.utc), False)

        # If user is negative then invalid code entered
        keypad_.setattr("code", code if user < 0 else "****", False)
        keypad_.setattr("last_user", user, True)

    def _ka_handler(self, keypad_areas: list[int]) -> None:
        for keypad in self.elements:
            if keypad_areas[keypad.index] >= 0:
                keypad.setattr("area", keypad_areas[keypad.index], True)

    def _kc_handler(self, keypad: int, key: int) -> None:
        keypads: list[Keypad] = cast(list[Keypad], self.elements)
        keypads[keypad].last_keypress = None  # Force a change notification
        try:
            name = KeypadKeys(key).name
        except ValueError:
            name = ""
        self.elements[keypad].setattr("last_keypress", (name, key), True)

    def _lw_handler(self, keypad_temps: list[int], zone_temps: list[int]) -> None:
        for keypad in self.elements:
            if keypad_temps[keypad.index] > -40:
                keypad.setattr("temperature", keypad_temps[keypad.index], True)

    def _st_handler(self, group: int, device: int, temperature: int) -> None:
        if group == 1:
            self.elements[device].setattr("temperature", temperature, True)



"""Definition of an ElkM1 Area"""

from __future__ import annotations

from typing import Any, cast

from .connection import Connection
from .const import ArmedStatus, Max, TextDescriptions
from .elements import Element, Elements
from .message import al_encode, as_encode, az_encode, dm_encode, zb_encode


class Area(Element):
    """Class representing an Area"""

    def __init__(self, index: int, connection: Connection) -> None:
        super().__init__(index, connection)
        self.armed_status: str | None = None
        self.arm_up_state: str | None = None
        self.alarm_state: str | None = None
        self.alarm_memory: str | None = None
        self.is_exit = False
        self.timer1 = 0
        self.timer2 = 0
        self.last_log: str | None = None

    def is_armed(self) -> bool:
        """Return if the area is armed."""
        return self.armed_status != ArmedStatus.DISARMED.value

    def arm(self, level: str, code: int) -> None:
        """(Helper) Arm system at specified level (away, vacation, etc)"""
        if self.is_armed() and level > "0":
            return
        self._connection.send(al_encode(level, self._index, code))

    def disarm(self, code: int) -> None:
        """(Helper) Disarm system."""
        self.arm("0", code)

    def display_message(
        self, clear: int, beep: bool, timeout: int, line1: str, line2: str
    ) -> None:
        """(Helper) Display a message on all of the keypads in this area."""
        self._connection.send(
            dm_encode(self._index, clear, beep, timeout, line1, line2)
        )

    def bypass(self, code: int) -> None:
        """(Helper) Bypass area."""
        self._connection.send(zb_encode(999, self._index, code))

    def clear_bypass(self, code: int) -> None:
        """(Helper) Clear bypass area."""
        self._connection.send(zb_encode(-1, self._index, code))


class Areas(Elements):
    """Handling for multiple areas"""

    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, Area, Max.AREAS.value)
        connection.msg_decode.add_handler("AM", self._am_handler)
        connection.msg_decode.add_handler("AS", self._as_handler)
        connection.msg_decode.add_handler("EE", self._ee_handler)
        connection.msg_decode.add_handler("LD", self._ld_handler)

    def sync(self) -> None:
        """Retrieve areas from ElkM1"""
        self._connection.send(as_encode())
        self.get_descriptions(TextDescriptions.AREA.value)

    def _am_handler(self, alarm_memory: str) -> None:
        for area in self.elements:
            area.setattr("alarm_memory", alarm_memory[area.index], True)

    def _as_handler(
        self, armed_statuses: str, arm_up_states: str, alarm_states: str
    ) -> None:
        update_alarm_triggers = False
        for area in self.elements:
            area = cast(Area, area)
            area.setattr("armed_status", armed_statuses[area.index], False)
            area.setattr("arm_up_state", arm_up_states[area.index], False)
            if (
                area.alarm_state != alarm_states[area.index]
                or alarm_states[area.index] != "0"
            ):
                update_alarm_triggers = True
            area.setattr("alarm_state", alarm_states[area.index], True)

        if update_alarm_triggers:
            self._connection.send(az_encode())

    def _ee_handler(
        self, msg_area: int, is_exit: bool, timer1: int, timer2: int, armed_status: str
    ) -> None:
        area = cast(Area, self.elements[msg_area])
        area.setattr("armed_status", armed_status, False)
        area.setattr("timer1", timer1, False)
        area.setattr("timer2", timer2, False)
        area.setattr("is_exit", is_exit, True)

    # ElkM1 global setting G35 must be set for LD messages to be sent
    def _ld_handler(self, area: int, log: dict[str, Any]) -> None:
        if log["event"] in [1173, 1174]:
            # arm/disarm log (YAGNI - decode number for more log types when needed)
            log["user_number"] = log["number"]
        self.elements[area].setattr("last_log", log, True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elkm1_lib']

package_data = \
{'': ['*']}

install_requires = \
['pyserial-asyncio>=0.5']

setup_kwargs = {
    'name': 'elkm1-lib',
    'version': '1.3.2',
    'description': 'Library for interacting with ElkM1 alarm/automation panel.',
    'long_description': '# Python ElkM1 library\n\n[![PyPI version](https://badge.fury.io/py/elkm1-lib.svg)](https://badge.fury.io/py/elkm1-lib)\n[![CI](https://github.com/gwww/elkm1/actions/workflows/code-quality.yml/badge.svg)](https://github.com/gwww/elkm1/actions/workflows/code-quality.yml)\n[![Downloads](https://pepy.tech/badge/elkm1-lib)](https://pepy.tech/project/elkm1-lib)\n[![License](https://img.shields.io/github/license/mashape/apistatus.svg)](https://pypi.org/project/elkm1_lib/)\n\nLibrary for interacting with ElkM1 alarm/automation panel.\n\nhttps://github.com/gwww/elkm1\n\n## Requirements\n\n- Python 3.6 (or higher)\n\n## Description\n\nThis package is created as a library to interact with an ElkM1 alarm/automation\npattern. The motivation to write this was to use with the Home Assistant\nautomation platform. The library can be used for writing other ElkM1 integration\napplications. The IO with the panel is asynchronous over TCP or over the\nserial port.\n\n## Installation\n\n```bash\n    $ pip install elkm1_lib\n```\n\n## Overview\n\nBasic connection to the Elk panel:\n\n```python\n    from elkm1_lib import Elk\n    import logging\n\n    # Print to STDOUT\n    LOG = logging.getLogger(__name__)\n    logging.basicConfig(level=logging.DEBUG, format=\'%(message)s\')\n\n    # Connect to elk\n    elk = Elk({\'url\': \'elk://192.168.1.100\'})\n    elk.connect()\n    elk.run()\n```\n\nThe above will connect to the Elk panel at IP address 192.168.1.100. the `elk://`\nprefix specifies that the connect is plaintext. Alternatively, `elks://` will\nconnect over TLS. In this case a userid and password must be specified\nand the call to `Elk` changes to:\n\n```python\n    elk = Elk(\n        {\'url\': \'elks://192.168.1.100\', \'userid\': \'test\', \'password\': \'pass\'}\n    )\n```\n\nThe following ElkM1 connection protocols are supported:\n- serial:// - Serial port;\n- elk:// - Elk M1XEP Ethernet, non-secure\n- elks:// - Elk M1XEP Ethernet, secure, TLS 1.0\n- elksv1_0:// - Elk M1XEP Ethernet, secure, TLS 1.0, supported on M1XEP version < 2.0.46\n- elksv1_2:// - Elk M1XEP Ethernet, secure, TLS 1.2, supported on M1XEP version = 2.0.46\n- elksv1_3:// - Elk M1XEP Ethernet, secure, TLS 1.2, not yet supported on M1XEP, reserved for future\n\nA username and password are required for any of the secure modes.\n\nTo see working example code take a look at the script `bin/simple`.\n\nThe `Elk` object supports the concept of `Elements`. An `Element`\nis the base class representation of `Zones`, `Lights`, etc. So, for\nexample there is a list of zones: `elk.zones` and each zone can be\naccessed by `elk.zones[index]`. Each element has a `__str__`\nrepresentation so that it is easy to print its contents.\n\nAll `Elements` are referenced starting at 0. Even though the Elk panel\nrefers to, for example, zones 1-208, the library references them\nas zones 0-207. All translation from base 0 to 1 and vice-versa is\nhandled internally in the `elkm1_lib.message` module.\n\nAfter creating the `Elk` object and connecting to the panel the\nlibrary code will synchronize all the elements to the data from the Elk panel.\n\nMany Elk messages are handled by the library, caching their contents. When a\nmessage causes a change to an attribute of an `Element`, registered\ncallbacks are called so that user use of the library can be notified\nof changing elements. The following user code shows registering a callback:\n\n```python\n    def call_me(element, changeset):\n       print(changeset)\n\n    for zone_number in range(Max.ZONES.value):\n      elk.zones[zone_number].add_callback(call_me)\n```\n\nThe library encodes, decodes, and processes messages to/from the\nElk panel. All the encoding and decoding is done in `elkm1_lib.message` module.\n\nMessages received are handled with callbacks. The library\ninternally registers callbacks so that decoded messages\ncan be used to update an `Element`. The user of the\nlibrary may also register callbacks. Any particular message\nmay have multiple callbacks.\n\nWhen the message is received it is decoded\nand some validation is done. The message handler is called\nwith the fields of from the decoded message. Each type of\nmessage has parameters that match the message type. All handler parameters\nare named parameters.\n\nHere is an example of a message handler being registered and how it is called:\n\n```python\n    def zone_status_change_handler(zone_number, zone_status):\n      print(zone_number, zone_status)\n\n    elk.add_handler(\'ZC\', zone_status_change_handler)\n```\n\nThe above code registers a callback for \'ZC\' (Elk zone status change)\nmessages. When a ZC message is received the handler functions are called\nwith the zone_number and zone_status.\n\nThere are a number of pseudo-handlers that act like the handlers. These are\ncalled when events happen. The pseudo-handlers are:\n\n- `connect`: When a successful connection to the ElkM1 is completed.\n- `disconnect`: When a connection to a panel is disconnected.\n- `login`: When a login is made to the panel (using `elks://` connection mode.\n  A single boolean parameter is passed `succeeded`.\n- `sync_complete`: When the panel has completed synchonizing all its elements.\n- `timeout`: When a send of a message to the ElkM1 times out (fails to send).\n- `unknown`: When a message from the ElkM1 is received and the library does\n  not have a method to decode the message. The message is passed to this handler\n  and can be decoded outside of the library.\n\n## Utilities\n\nThe `bin` directory of the library has one utility program and\na couple of example uses of the library.\n\n### `mkdoc`\n\nThe utility `mkdoc` creates a Markdown table of the list of Elk\nmessages with a check mark for those messages have encoders/decoders\nand an X for those messages are not planned to be implemented.\nThere are no parameters to `mkdoc`. It outputs to stdout.\nThe data for the report comes from the ElkM1 library code mostly.\nA couple of things are hard coded in the mkdoc script, notably\nthe "no plans to implement" list.\n\n### `simple`\n\nThe `simple` Python script is a trivial use of the ElkM1 library.\nIt connects to the panel, syncs to internal memory, and continues\nlistening for any messages from the panel. The URL of the ElkM1 to\nconnect to is retrieved from an environment variable named `ELKM1_URL`.\n\n### `elk`\n\nThe `elk` Python script is a bit of a command interpretor. It can run in\ntwo modes. Non-interactive mode is the default. Just run the `elk` command.\nThe non-interactive mode is similar to `simple` except there are a\ncouple of message handlers (`timeout` and `unknown` handlers).\n\nThe `elk` can also be run in interactive mode by invoking it by\n`elk -i`. In this mode is uses curses (full screen use of the terminal)\nthat has a command line and an output window. `TAB` switches between\nthe command line and output windows. In the output window the arrow keys\nand scrollwheel scroll the contents of the window.\n\nIn the command line when running `elk -i` there are a\nnumber of commands. Start with `help`. Then `help <command>` for\ndetails on each command. In general there are commands to dump the internal\nstate of elements and to invoke any of the encoders to send a message\nto the Elk panel.\n\nFor example, `light <4, 8, 12-14` would invoke the `__str__` method\nfor the light element to print the cached info for lights 0-3, 8, and 12-14.\n\nAnother example would be `pf 3` which issues the pf (Turn light off)\ncommand for light number 3 (light 4 on the panel -- remember 0\nversus 1 base).\n\nAll of the commands that send messages to the panel are automatically\ndiscovered and are all the XX_encode functions in the ``elkm1_lib.message``\nmodule. The docstring and the XX_encode\'s parameters are shown as part\nof the help.\n\n## Development\n\nThis project uses [poetry](https://poetry.eustace.io/) for development dependencies. Installation instructions are on their website.\n\nTo get started developing:\n\n```\ngit clone https://github.com/gwww/elkm1.git\ncd elkm1\npoetry install\npoetry shell # Or activate the created virtual environment\nmake test # to ensure everything installed properly\n```\n\nThere is a `Makefile` in the root directory as well. The `make` command\nfollowed by one of the targets in the `Makefile` can be used. If you don\'t\nhave or wish to use `make` the `Makefile` serves as examples of common\ncommands that can be run.\n\n## Reporting a Bug\n\nNo problem ;) — report the bugs! But, logs are most often required. If you\nare using Home Assistant, which is about the only use I\'m aware of for\nthis library, then add the following to your `configuration.yaml`:\n```\nlogger:\n  default: info\n  logs:\n    custom_components.elkm1: debug\n    elkm1_lib: debug\n```\n\nDo everything in your power to trim to logs down to their smallest. One way is\nto reproduce your problem quickly so that few other logs are not generated in\nbetween. Another recommendation is to use the simplest configuration that you\ncan think of to reproduce the problem.\n\nCan you reproduce the problem in other ways? If this is a problem that is\nbeing experienced while using Home Assistant try using the `Services` in `Developer Tools`.\n\nSometime logs may have sensitive information in them. You may want to\nscan your logs for that info and "X" it out. In addition, you can send logs\ndirectly to me. Support email is in the `pyproject.toml` file. You may also\nsend a link to somewhere you\'ve stored/shared the file (DropBox for example).\n',
    'author': 'Glenn Waters',
    'author_email': 'gwwaters+elkm1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gwww/elkm1',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
